#include <iostream>
#include <iomanip>

extern "C"
{
#include <usb.h>
}

#include <usbdriver.h>

using namespace std;
#define USB_TIMEOUT 500

// EEPROM Read command for Microchip 25AA02 SPI EEPROMs
#define EEPROM_READ_CMD 0x03

usb_dev_handle * hTinyISP;


usb_dev_handle * acquireTinyISP(void)
{
  struct usb_bus * thisBus;
  struct usb_device * thisDevice = NULL;

  usb_init();
  usb_find_busses();
  usb_find_devices();

  for (int retry = 0; retry < 1 && NULL == thisDevice; retry++)
    {

      for (thisBus = usb_get_busses(); NULL != thisBus; thisBus = thisBus->next)
	{
	  
	  for (thisDevice = thisBus->devices; NULL != thisDevice; thisDevice = thisDevice->next)
	    {
	      usb_dev_handle * device;
	      device = usb_open(thisDevice);
	      //cout << "Vendor " << hex << setw(4) << setfill('0')
	      //   << thisDevice->descriptor.idVendor
	      //   << " Product " << thisDevice->descriptor.idProduct
	      //   << endl;

	      if (USBTINY_VENDOR == thisDevice->descriptor.idVendor &&
		  USBTINY_PRODUCT == thisDevice->descriptor.idProduct)
		{
		  cout << "Acquired device" << endl;
		  return device;
		}
	      usb_close(device);
	      
	    }
	}
      cerr << "No device found, sleeping..." << endl;
      sleep(1);
    }
}

void usb_control ( int req, int val, int size )
{
  usb_control_msg( hTinyISP,
		   USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		   req, val, size, NULL, 0, USB_TIMEOUT );
}

// Wrapper for simple usb_control_msg messages to receive data from
// programmer. This is taken from avrdude v5.8 and modified slightly
// to fit in with the original usbdriver.cpp code.
static int usb_in (unsigned int requestid, unsigned int val, unsigned int index,
		   unsigned char* buffer, int buflen)
{
  int nbytes;
  int i;

  for (i = 0; i < 10; i++) {
    nbytes = usb_control_msg( hTinyISP,
			      USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			      requestid,
			      val, index,
			      (char *)buffer, buflen,
			      USB_TIMEOUT);
    if (nbytes == buflen) {
      return nbytes;
    }
  }
  fprintf(stderr, "\n%s: error: usbtiny_receive: %s (expected %d, got %d)\n",
          "usbdriver", usb_strerror(), buflen, nbytes);
  return -1;
}

void readFromEEPROM(string data)
{
  unsigned int w1, w2; // w1 and w2 need to be at least 16 bits each
  unsigned char res[4]; 
  int offsetConv;
  unsigned char offset;
  int i;

  // handle input argument and convert it from a hexadecimal value to
  // a useable unsigned integer.
  {
    // convert input argument to all uppercase
    for (i = 0; i < data.length(); i++) { 
      data[i]=toupper(data[i]);
    }

    // remove leading "0"
    while (data[0] == '0') {
      data.erase(0, 1);
    }

    // remove any x (.ie number began with "0x")
    // remember I made it uppercase so that case does not matter
    if (data[0] == 'X') {
      data.erase(0, 1);
    }

    // Use sscanf to convert the remaining hex digits
    sscanf(data.c_str(), "%X", &offsetConv);
    offset = offsetConv & 0xff; // only use low byte
  }

  // Send EEPROM Read Command and entered byte offset from command
  // line argument.
  w1 = EEPROM_READ_CMD + (offset << 8);

  // send two dummy bytes to complete transfer of 4 bytes and to read
  // out the result.
  w2 = 0x00 + (0x00 << 8);

  // NOTE: If wanted to manually force the RESET or /SS line low, can do
  // it with the below line.  However, this is not needed because the
  // POWERUP command that was sent earlier lowered the /SS line.
  //
  // usb_control(USBTINY_CLR, 4, 1); // /SS low (PB4)

  usb_in (USBTINY_SPI, w1, w2, res, sizeof(res));

  // NOTE: If wanted to manually force the RESET or /SS line high
  // after the transfer, can do it with the below line.  However, this
  // is not needed because the POWERDOWN command that will be sent
  // next will raise the /SS line.
  //
  // usb_control(USBTINY_SET, 4, 1); // /SS high (PB4)

  // byte at offset gets returned in the third byte 
  //
  // NOTE: the byte after the offset is returned in the fourth byte but
  // we only care about the byte for the requested offset.
  printf("EEPROM byte at offset 0x%02x: 0x%02x\n", 
	 offset, res[2]);

}

// NOTE: The dataToDisplay() function is left over from the original
// code.  It is never called by this revised code.
void dataToDisplay(string data)
{
  cout << "Display " << data.c_str() << endl;

  usb_control(USBTINY_CLR, 4, 1); // /SS low (PB4)
  for (string::iterator iData = data.begin();
       iData != data.end();
       iData++)
    {
      usb_control(USBTINY_SPI1, *iData,  1);	// data
    }
  usb_control(USBTINY_SET, 4, 1); // /SS high (PB4)
}

int main(int argc, char ** argv)
{
  if (argc != 2) {
    fprintf(stderr, "\nERROR: Must enter the desired byte offset in hexadecimal format, eg. \n\n%s 0xfa\n\n\n",
	    argv[0]);
  } else {

    hTinyISP = acquireTinyISP();

    if (NULL == hTinyISP) {
      cerr << "Could not acquire device" << endl;
      exit(1);
    }

    // get ready for transfer and force RESET low.  The RESET line in
    // AVR speak is the /SS or /CS line for SPI.  This needs to be low
    // in order to enable the SPI device for access.
    usb_control( USBTINY_POWERUP, USB_TIMEOUT, RESET_LOW );

    readFromEEPROM(argv[1]);

    usb_control( USBTINY_POWERDOWN, 0, 0 );
    usb_close(hTinyISP);
  }

}
